from sqlalchemy import Integer, Text, ForeignKey, DateTime, func, Boolean
from datetime import datetime
from sqlalchemy.orm import Mapped, mapped_column
from .base import Base


class DraftEvent(Base):
    __tablename__ = "draft_events"
    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    game_id: Mapped[str] = mapped_column(Text, ForeignKey("games.id"))
    champion_id: Mapped[int] = mapped_column(Integer, ForeignKey("champions.id"))
    # Derived from game_info based on champion_id
    participant_id: Mapped[int | None] = mapped_column(Integer, nullable=True)
    # True if pick event, false if ban event
    is_pick: Mapped[bool] = mapped_column(Boolean)
    # True if first 3 bans/picks, False if last 2 picks/bans
    is_phase_one: Mapped[bool] = mapped_column(Boolean)
    # True if blue team, false if red team (100 / 200)
    is_blue: Mapped[bool] = mapped_column(Boolean)
    # Picks: blue1 red2, red3, blue4, blue5, red6, red7, blue8, blue9, red10
    # Bans: 
    #     Phase 1: 1-6
    #     Phase 2: 1-4
    pick_turn: Mapped[int] = mapped_column(Integer)
    # Debug
    updated: Mapped[datetime] = mapped_column(DateTime, default=func.now(), onupdate=func.now())



import requests as r
from requests import Response
from .utils import headers, RegionLiteral, QueueLiteral, DivisionLiteral, TierLiteral
from ..rate_limiter import riot_api_limiter


@riot_api_limiter(endpoint_key="league_v4_get_challengers")
def get_challengers(
    region: RegionLiteral, queue: QueueLiteral, api_key: str
) -> Response:
    _headers = {"X-Riot-Token": api_key, **headers}
    response = r.get(
        f"https://{region.lower()}.api.riotgames.com/lol/league/v4/challengerleagues/by-queue/{queue}",
        headers=_headers,
    )
    return response


@riot_api_limiter(endpoint_key="league_v4_get_grandmasters")
def get_grandmasters(
    region: RegionLiteral, queue: QueueLiteral, api_key: str
) -> Response:
    _headers = {"X-Riot-Token": api_key, **headers}
    response = r.get(
        f"https://{region.lower()}.api.riotgames.com/lol/league/v4/grandmasterleagues/by-queue/{queue}",
        headers=_headers,
    )
    return response


@riot_api_limiter(endpoint_key="league_v4_get_masters")
def get_masters(region: RegionLiteral, queue: QueueLiteral, api_key: str) -> Response:
    _headers = {"X-Riot-Token": api_key, **headers}
    response = r.get(
        f"https://{region.lower()}.api.riotgames.com/lol/league/v4/masterleagues/by-queue/{queue}",
        headers=_headers,
    )
    return response


@riot_api_limiter(endpoint_key="league_v4_entries_puuid")
def get_league_entries_by_puuid(
    region: RegionLiteral, puuid: str, api_key: str
) -> Response:
    _headers = {"X-Riot-Token": api_key, **headers}
    response = r.get(
        f"https://{region.lower()}.api.riotgames.com/lol/league/v4/entries/by-puuid/{puuid}",
        headers=_headers,
    )
    return response


@riot_api_limiter(endpoint_key="league_v4_entries_summoner_id")
def get_league_entries_by_summoner_id(
    region: RegionLiteral, summoner_id: str, api_key: str
) -> Response:
    _headers = {"X-Riot-Token": api_key, **headers}
    response = r.get(
        f"https://{region.lower()}.api.riotgames.com/lol/league/v4/entries/by-summoner/{summoner_id}",
        headers=_headers,
    )
    return response


@riot_api_limiter(endpoint_key="league_v4_all_entries")
def get_all_league_entries(
    region: RegionLiteral,
    queue: QueueLiteral,
    tier: TierLiteral,
    division: DivisionLiteral,
    api_key: str,
    page: int = 1,
) -> Response:
    _headers = {"X-Riot-Token": api_key, "page": str(page), **headers}
    response = r.get(
        f"https://{region.lower()}.api.riotgames.com/lol/league/v4/entries/{queue}/{tier}/{division}",
        headers=_headers,
    )
    return response


@riot_api_limiter(endpoint_key="league_v4_entries_league_id")
def get_league_entries_by_league_id(
    region: RegionLiteral, league_id: str, api_key: str
) -> Response:
    _headers = {"X-Riot-Token": api_key, **headers}
    response = r.get(
        f"https://{region.lower()}.api.riotgames.com/lol/league/v4/leagues/{league_id}",
        headers=_headers,
    )
    return response



from sqlalchemy import Integer, Text, ForeignKey
from sqlalchemy.orm import relationship, Mapped, mapped_column
from .base import Base


class PlayerTeamAssociation(Base):
    __tablename__ = "player_team_associations"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    player_id: Mapped[int] = mapped_column(Integer, ForeignKey("players.id"))
    team_id: Mapped[str] = mapped_column(Text, ForeignKey("teams.id"))
    # Although not explicitly enforced, the (TOP, JUNGLE, MIDDLE, BOTTOM, UTILITY) is generally used
    position: Mapped[str | None] = mapped_column(Text, nullable=True)

    team: Mapped["Team"] = relationship("Team", back_populates="player_associations")

    def __repr__(self):
        return f"<PlayerTeamAssociation(id='{self.id}', player_id='{self.player_id}', team_id='{self.team_id}', position='{self.position}')>"



from .constants import (
    Side,
    Lane,
    TeamPosition,
    DEFAULT_REGION,
    SEASON_START,
    MINIMUM_MATCH_DURATION,
    SeriesStatus,
    SeriesType
)

from .helpers import camel_to_snake, snake_to_camel

__all__ = [
    "Side",
    "Lane",
    "TeamPosition",
    "DEFAULT_REGION",
    "SEASON_START",
    "MINIMUM_MATCH_DURATION",
    "SeriesStatus",
    "SeriesType",
    "camel_to_snake",
    "snake_to_camel",
]



from sqlalchemy import (
    ForeignKey,
    Text,
    Integer,
    BigInteger,
    DateTime,
    func,
    Index,
    text,
    Boolean,
    case,
    literal
)
from sqlalchemy.orm import Mapped, mapped_column, relationship, MappedColumn
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.dialects.postgresql import JSONB
from datetime import datetime
from .base import Base
from ..api.utils import REGIONS


class Game(Base):
    __tablename__ = "games"
    # InfoDto
    end_of_game_result: Mapped[str | None] = mapped_column(Text, nullable=True)
    game_duration: Mapped[int | None] = mapped_column(
        Integer, nullable=True
    )  # in seconds
    game_id: Mapped[int | None] = mapped_column(
        BigInteger, nullable=True
    )  # Riot's game id
    game_mode: Mapped[str | None] = mapped_column(Text, nullable=True)
    game_name: Mapped[str | None] = mapped_column(Text, nullable=True)
    game_type: Mapped[str | None] = mapped_column(Text, nullable=True)
    game_version: Mapped[str | None] = mapped_column(
        Text, nullable=True
    )  # Riot game version
    map_id: Mapped[int | None] = mapped_column(Integer, nullable=True)
    platform_id: Mapped[str | None] = mapped_column(Text, nullable=True)  # e.g. EUW1
    queue_id: Mapped[int | None] = mapped_column(Integer, nullable=True)
    tournament_code: Mapped[str | None] = mapped_column(Text, nullable=True)

    INFO_DTO = [
        name for name, value in locals().items() if isinstance(value, MappedColumn)
    ]

    game_creation: Mapped[datetime | None] = mapped_column(DateTime(timezone=True), nullable=True)
    game_start_timestamp: Mapped[datetime | None] = mapped_column(DateTime(timezone=True), nullable=True)
    game_end_timestamp: Mapped[datetime | None] = mapped_column(DateTime(timezone=True), nullable=True)

    TIMESTAMPS = [
        name for name, value in locals().items() if isinstance(value, MappedColumn) and isinstance(value.column.type, DateTime)
    ]

    # Equivalent to matchId in the MatchV5 API (e.x. NA1_12345)
    id: Mapped[str] = mapped_column(Text, primary_key=True)

    game_ended_in_early_surrender: Mapped[bool | None] = mapped_column(Boolean)
    game_ended_in_surrender: Mapped[bool | None] = mapped_column(Boolean)

    # Additional game information can be stored here.
    source_data = mapped_column(JSONB)

    # Esports Game Information
    series_id: Mapped[str | None] = mapped_column(Text, ForeignKey("series.id"), nullable=True)
    series_game_number: Mapped[int | None] = mapped_column(Integer, nullable=True)

    # Debug
    updated: Mapped[datetime] = mapped_column(DateTime, default=func.now(), onupdate=func.now())

    # ParticipantDto
    participants: Mapped[list["Participant"]] = relationship(
        "Participant", back_populates="game"
    )
    # TeamDto
    teams: Mapped[list["TeamDto"]] = relationship()

    @hybrid_property
    def patch(self):  # type: ignore
        version_split = self.game_version.split(".")
        return (
            f"{version_split[0]}.{version_split[1]}"
            if len(version_split) >= 2
            else None
        )

    @patch.expression
    def patch(cls):
        return func.concat(
            func.split_part(cls.game_version, ".", 1),
            ".",
            func.split_part(cls.game_version, ".", 2),
        )

    @hybrid_property
    def solo_queue(self) -> bool: # type: ignore
        return self.platform_id in list(REGIONS)

    @solo_queue.expression
    def solo_queue(cls):
        return cls.platform_id.in_(list(REGIONS))

    @hybrid_property
    def calculated_game_type(self) -> str: # type: ignore
        if self.solo_queue and self.queue_id == 420:
            return "SOLOQUEUE"
        elif self.solo_queue and self.queue_id == 0: # Tournament code games
            return "CUSTOM"
        elif not self.solo_queue and self.game_name and self.game_name.lower().startswith("scrim|"):
            return "SCRIM"
        elif not self.solo_queue:
            return "ESPORTS"
        else:
            return ""

    @calculated_game_type.expression
    @classmethod
    def calculated_game_type(cls):
        return case(
            (cls.solo_queue.is_(True) & (cls.queue_id == 420), literal("SOLOQUEUE")),
            (cls.solo_queue.is_(True) & (cls.queue_id == 0), literal("CUSTOM")),
            (cls.solo_queue.is_(False) & cls.game_name.ilike("scrim|%"), literal("SCRIM")),
            (cls.solo_queue.is_(False), literal("ESPORTS")),
            else_=literal("")
        )

    __table_args__ = (
        Index("idx_games_queue_id", "queue_id"),
        Index("idx_games_end_timestamp", text("game_end_timestamp DESC")),
        Index("idx_game_queue_timestamp", "queue_id", text("game_end_timestamp DESC")),
    )



from datetime import timedelta
from enum import Enum, StrEnum, auto

DEFAULT_REGION = "NA"
SEASON_START = 1736420400  # Season 15. Season 14 was 1704884400
MINIMUM_MATCH_DURATION = timedelta(minutes=10)

class SeriesType(StrEnum):
    """Enum to represent how the series types are represented in the GRID API
    
    This originally was found in central_data.enums's SeriesType"""
    COMPETITIVE = auto()
    ESPORTS = auto()
    LOOPFEED = auto()
    SCRIM = auto()

class SeriesStatus(StrEnum):
    """Enum to represent many of the possible states that an esports series may be in.
    
    Probably won't use all of them immediately but this should serve as a pretty good basis.
    """
    ###    Pre-game states    ###
    SCHEDULED = auto()          # Series is scheduled but hasn't started yet   
    CANCELLED = auto()          # Series was cancelled
    
    ###   During-game states  ###
    IN_PROGRESS = auto()        # Series is currently being played
    PAUSED = auto()             # Series is temporarily paused (technical issues, etc.)
    
    ###    Post-game states   ###
    PENDING_PROCESSING = auto() # Series complete, waiting for processing
    PROCESSING = auto()         # Currently being processed (processing lock)
    PARTIAL = auto()            # Some games ingested successfully, others missing
    INCOMPLETE = auto()         # Some data is missing (players, events, etc.)
    
    ###      Final states     ###
    VALID = auto()              # All games and data are ingested and valid
    FAILED = auto()             # Processing failed completely, needs retry
    BUGGED = auto()             # Data exists but has inconsistencies/errors or failed on retry
    
    ###      Special cases    ###
    NO_DATA_AVAILABLE = auto()  # Series exists in GRID but no game data available
    NEEDS_REVIEW = auto()       # Flagged for manual review
    

class Side(Enum):
    blue = 100
    red = 200


class Lane(Enum):
    TOP = "top_lane"
    JNG = "jungle"
    MID = "mid_lane"
    BOT = "bot_lane"
    SUP = "utility"


class TeamPosition(Enum):
    AFK = None
    BOTTOM = "BOT"
    JUNGLE = "JNG"
    MIDDLE = "MID"
    TOP = "TOP"
    UTILITY = "SUP"



from . import api
from . import database
from . import match_lib
from . import models
from . import rate_limiter
from . import utils

__all__ = ["api", "database", "match_lib", "models", "rate_limiter", "utils"]



from sqlalchemy import Integer, Text, ForeignKey
from sqlalchemy.orm import Mapped, mapped_column
from .base import Base


class TeamStat(Base):
    __tablename__ = "team_stats"
    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    game_id: Mapped[str] = mapped_column(Text, ForeignKey("games.id"))
    team_id: Mapped[int] = mapped_column(Integer) # 100 / 200
    game_time: Mapped[int] = mapped_column(Integer)

    # Team stats from stat_update
    deaths: Mapped[int] = mapped_column(Integer)
    assists: Mapped[int] = mapped_column(Integer)
    champion_kills: Mapped[int] = mapped_column(Integer)

    total_gold: Mapped[int] = mapped_column(Integer)
    baron_kills: Mapped[int] = mapped_column(Integer)
    inhib_kills: Mapped[int] = mapped_column(Integer)
    tower_kills: Mapped[int] = mapped_column(Integer)
    dragon_kills: Mapped[int] = mapped_column(Integer)

    """

    'deaths': 8,
       'teamID': 100,
       'assists': 54,
       'totalGold': 55997,
       'baronKills': 1,
       'inhibKills': 1,
       'towerKills': 7,
       'dragonKills': 3,
       'championsKills': 21},"""



from .base import Base
from .account import Account
from .champion import Champion
from .draft_event import DraftEvent
from .game import Game
from .game_event import GameEvent
from .participant_stat import ParticipantStat
from .participant import Participant
from .player import Player
from .series import Series
from .player_team_association import PlayerTeamAssociation
from .team import Team
from .team_dto import TeamDto
from .team_stat import TeamStat
from .tournament import Tournament

__all__ = [
    "Base",
    "Account",
    "Champion",
    "DraftEvent",
    "Game",
    "GameEvent",
    "Player",
    "Series",
    "ParticipantStat",
    "Participant",
    "PlayerTeamAssociation",
    "Team",
    "TeamDto",
    "TeamStat",
    "Tournament",
]



from sqlalchemy import Integer, Text, DateTime, func
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import relationship, Mapped, mapped_column
from .base import Base
from .account import Account
from datetime import datetime


class Player(Base):
    __tablename__ = "players"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    # Not sure how to deal with naming conflicts. For now, we assume that names can be repeated but we can use the column below.
    name: Mapped[str] = mapped_column(Text)
    # We store the name used on Leaguepedia/Oracles Elixer here for easier use with those services.
    # For now, this should be their Leaguepedia disambiguation name
    disambiguation: Mapped[str | None] = mapped_column(Text, unique=True)
    # We store linked IDs (GRID, RIOT_ESPORTS, Discord) in a flexible format
    external_ids = mapped_column(JSONB)
    # Debug
    updated: Mapped[datetime] = mapped_column(DateTime, default=func.now(), onupdate=func.now())

    accounts: Mapped[list["Account"]] = relationship("Account", back_populates="player")

    def __repr__(self) -> str:
        if self.disambiguation is None:
            return f"{self.id}-{self.name}:{self.id}"
        return f"{self.id}-{self.name} ({self.disambiguation})"



from typing import List
from sqlalchemy import select, update
from ATG.models import Player, Account
from ATG.api import get_account_by_riot_id

def link_pro(session, pro_name, soloq, RIOT_API, inactive=False, region="NA1"):
    try:
        player = session.execute(select(Player).where(Player.name == pro_name)).scalar_one()
    except:
        print(f"Unable to find associated player for {pro_name}")
        return
    name, tagline = soloq.split("#")
    details = get_account_by_riot_id(name, tagline, RIOT_API).json()
    new_acc = Account(puuid=details['puuid'], name=details['gameName'], tagline=details['tagLine'], region=region, player_id=player.id, skip_update = inactive)
    session.add(new_acc)
    try:
        session.commit()
        print(f"Linked {pro_name} with {name}#{tagline}")
    except:
        session.rollback()
        print(f"Account {details['gameName']}#{details['tagLine']} already linked")
def list_accounts(session, pro_name):
    try:
        player = session.execute(select(Player).where(Player.name == pro_name)).scalar_one()
    except:
        print(f"Unable to find associated player for {pro_name}")
        return
    print(["#".join(_) for _ in list(session.execute(select(Account.name, Account.tagline).where(Account.player_id == player.id)).all())])

def create_player(session, pro_name):
    player = session.execute(select(Player).where(Player.name == pro_name)).scalar_one_or_none()
    if player is not None:
        print(f"Player with name {player.name} already exists")
        return
    new_player = Player(name=pro_name)
    session.add(new_player)
    session.commit()
    print(f"Created new player {new_player.name} with id {new_player.id}")



from sqlalchemy import Integer, Text, ForeignKey, Boolean, DateTime, func
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import Mapped, mapped_column, relationship
from datetime import datetime
from .base import Base


class TeamDto(Base):
    __tablename__ = "team_dtos"
    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    game_id: Mapped[str] = mapped_column(Text, ForeignKey("games.id"))

    # TeamDto
    bans = mapped_column(JSONB)
    objectives = mapped_column(JSONB)
    team_id: Mapped[int] = mapped_column(Integer)
    win: Mapped[bool] = mapped_column(Boolean)

    fk_team_id: Mapped[str | None] = mapped_column(Text, ForeignKey("teams.id"), nullable=True)

    # Debug
    updated: Mapped[datetime] = mapped_column(DateTime, default=func.now(), onupdate=func.now())

    participants: Mapped[list["Participant"]] = relationship(
        "Participant",
        primaryjoin="and_(TeamDto.game_id == Participant.game_id, TeamDto.team_id == Participant.team_id)",
        foreign_keys="[Participant.game_id, Participant.team_id]",
        viewonly=True,
    )



from sqlalchemy import Text, DateTime, func
from sqlalchemy.orm import relationship, Mapped, mapped_column
from sqlalchemy.dialects.postgresql import JSONB
from datetime import datetime

from .base import Base
from .player_team_association import PlayerTeamAssociation


class Team(Base):
    __tablename__ = "teams"
    id: Mapped[str] = mapped_column(Text, primary_key=True)
    # Multiple teams can share the same name in the GRID and LoL Esports APIs
    name: Mapped[str] = mapped_column(Text, unique=False)
    # Multiple teams can share the same team code due to inconsistancies in GRID/Bayes data e.x. academy teams share team codes with their main team
    team_code: Mapped[str | None] = mapped_column(Text, unique=False, nullable=True)
    # We can store GRID, LOL IDs here
    external_ids = mapped_column(JSONB)

    # We store additional information in a JSONB blob
    additional_details = mapped_column(JSONB)

    # Debug
    updated: Mapped[datetime] = mapped_column(DateTime, default=func.now(), onupdate=func.now())

    player_associations: Mapped[list["PlayerTeamAssociation"]] = relationship(
        "PlayerTeamAssociation", back_populates="team"
    )

    def __repr__(self):
        return f"<Team(id='{self.id}', code='{self.name}')>"



import requests as r
from requests import Response
from .utils import headers, PLATFORM_ROUTING, parse_match_id, RegionLiteral
from ..rate_limiter import riot_api_limiter


@riot_api_limiter(endpoint_key="match_v5_by_id")
def get_match_by_id(match_id: str, api_key: str, timeline: bool = False) -> Response:
    region, _ = parse_match_id(match_id)
    is_timeline = "/timeline" if timeline else ""
    _headers = {"X-Riot-Token": api_key, **headers}
    response = r.get(
        f"https://{PLATFORM_ROUTING[region]}.api.riotgames.com/lol/match/v5/matches/{match_id}{is_timeline}",
        headers=_headers,
    )
    return response


@riot_api_limiter(endpoint_key="match_v5_available_matches")
def get_available_matches(
    puuid: str, region: RegionLiteral, api_key: str, **kwargs
) -> Response:
    _headers = {"X-Riot-Token": api_key, **headers}
    response = r.get(
        f"https://{PLATFORM_ROUTING[region]}.api.riotgames.com/lol/match/v5/matches/by-puuid/{puuid}/ids",
        headers=_headers,
        params=kwargs,
    )
    return response


def get_match_history(
    puuid: str, region: RegionLiteral, api_key: str, **kwargs
) -> list[str]:
    start = 0
    count = 100
    match_history = []
    while True:
        match_list = get_available_matches(
            puuid=puuid,
            region=region,
            api_key=api_key,
            start=start,
            count=count,
            **kwargs,
        )
        if match_list:
            try:
                match_list.raise_for_status()
                match_list = match_list.json()
                if len(match_list) == 0:
                    break
                match_history.extend(match_list)
                start += count
            except:
                # If an error occurs, we will fail.
                print(f"{match_list.status_code} - Failed to retrieve match history")
                return match_history
    return match_history



from datetime import datetime
from sqlalchemy import Text, Integer, DateTime, func
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.types import SmallInteger
from sqlalchemy.orm import Mapped, mapped_column, relationship
from .base import Base


class Tournament(Base):
    __tablename__ = "tournaments"
    id: Mapped[str] = mapped_column(Text, primary_key=True)
    # Tournament Name
    name: Mapped[str] = mapped_column(Text)
    # We should automatically parse the tournament year / league / split from its name for easy searching
    year: Mapped[int | None] = mapped_column(SmallInteger, nullable=True)
    league: Mapped[str | None] = mapped_column(Text, nullable=True)
    split: Mapped[str | None] = mapped_column(Text, nullable=True)
    # Similar to the playoffs boolean from OracleElixer, however we can store the
    # additional tournament tag grid gives us inside brackets
    event_type: Mapped[str | None] = mapped_column(Text, nullable=True)

    # For now, these are generally NULL, in the future we should automatically set based on first/last games
    start_date: Mapped[datetime | None] = mapped_column(DateTime, nullable=True)
    end_date: Mapped[datetime | None] = mapped_column(DateTime, nullable=True)

    # External IDs can be stored here
    external_ids = mapped_column(JSONB)

    # Additional team information can be stored here 
    source_data = mapped_column(JSONB)
    
    # Debug
    updated: Mapped[datetime] = mapped_column(DateTime, default=func.now(), onupdate=func.now())



from ..models import Player, Account, Participant
from sqlalchemy.orm import Session
from sqlalchemy import select

def create_player(name: str) -> Player:
    """Creates a player object

    Args:
        name (str): Player name

    Returns:
        Player: Created player object
    """
    return Player(name=name)

def create_account(puuid: str, name: str, tagline: str, player_id: int, region: str = "TOURNAMENT", skip_update: bool = True) -> Account:
    """Creates an account object

    Args:
        puuid (str): Account identifier (PUUID / Account ID)
        name (str): Account name
        tagline (str): Account tagline
        player_id (int): Player ID (foreign key)
        region (str, optional): Account region. Defaults to "TOURNAMENT".
        skip_update (bool, optional): Should be updated automatically?. Defaults to True.

    Returns:
        Account: Created account object
    """
    return Account(
        puuid=puuid,
        name=name,
        tagline=tagline,
        region=region,
        player_id=player_id,
        skip_update=skip_update)

def get_puuid(participant) -> str:
    return str(participant.get("puuid", participant["accountID"]))
def get_game_name(participant) -> str:
    return participant.get("riotId", {"displayName": participant.get("summonerName")})["displayName"]
def get_tagline(participant) -> str:
    return participant.get("riotId", {}).get("tagLine", "")
def get_team_id(participant) -> int:
    return participant.get("teamID", participant.get("teamId"))

def parse_game_participants(participants: list[dict], game_id: str, winning_team_id: int, NAME_ID_MAP: dict[str, int]) -> list[Participant]:
    """Parses the participants into a list of Participant objects

    Args:
        participants (list[dict]): List of game participants
        winning_team_id (int): 100 or 200 (blue/red winners)
        NAME_ID_MAP (dict[str, int]): champion name -> champion id map

    Returns:
        list[Participant]: Parsed participant objects
    """
    game_participants = []
    for participant in participants:
        game_participants.append(Participant(
            game_id = game_id,
            puuid = get_puuid(participant),
            participant_id = participant["participantID"],
            riot_id_game_name = get_game_name(participant),
            riot_id_tagline = get_tagline(participant),
            summoner_id = participant["accountID"],
            team_id = get_team_id(participant),
            champion_id = NAME_ID_MAP[participant["championName"].lower()],
            win = get_team_id(participant) == winning_team_id,
        ))
    return game_participants

def get_common_prefix_length(player_names: list[str]) -> int:
    """Gets the number of charaters shared in the prefix of all items of a list

    Args:
        player_names (list[str]): A list of player names

    Returns:
        int: The length of the shared prefix
    """
    common_prefix_length = 0
    for chars in zip(*player_names):
        if len(set(chars)) == 1:
            common_prefix_length += 1
        else:
            break
    return common_prefix_length

def create_missing_accounts(session: Session, participants: dict):
    """Helper function to parse a list of participants and add to the session any missing players/accounts.

    Args:
        session (Session): Active SQLAlchemy session object
        participants (dict): A match-v5 like participant list
    """
    stored_puuids = session.scalars(select(Account.puuid)).all()
    for participant in participants:
        puuid = get_puuid(participant)
        if puuid not in stored_puuids:
            game_name = get_game_name(participant)
            player_names = [get_game_name(p) for p in participants if get_team_id(p) == get_team_id(participant)]
            player_name = game_name[get_common_prefix_length(player_names):]
            
            db_player = session.scalar(select(Player).where(Player.name == player_name))
            
            if db_player is None:
                db_player = create_player(player_name)
                session.add(db_player)
                session.flush()

            session.add(create_account(puuid, game_name, get_tagline(participant), db_player.id))


from sqlalchemy import ForeignKey, Text, Integer, DateTime, func, Index, Boolean
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import Mapped, mapped_column, MappedColumn, relationship
from datetime import datetime
from .base import Base


class Participant(Base):
    __tablename__ = "participants"
    # ParticipantDto
    # An exception to the Obj_id foreign key naming is participant_id which referes to the relative ID
    # of the participant to other participants in a game.
    participant_id: Mapped[int] = mapped_column(Integer)
    champion_id: Mapped[int] = mapped_column(Integer, ForeignKey("champions.id"))
    puuid: Mapped[str | None] = mapped_column(Text)
    riot_id_game_name: Mapped[str] = mapped_column(Text)
    riot_id_tagline: Mapped[str] = mapped_column(Text)
    summoner_id: Mapped[str | None] = mapped_column(Text)
    summoner_name: Mapped[str | None] = mapped_column(Text)
    team_position: Mapped[str | None] = mapped_column(Text)
    team_id: Mapped[int] = mapped_column(Integer)  # RENAME
    win: Mapped[bool | None] = mapped_column(Boolean)

    # Automatically generate the stored_keys
    PARTICIPANT_DTO = [
        name for name, value in locals().items() if isinstance(value, MappedColumn)
    ]

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    game_id: Mapped[str] = mapped_column(Text, ForeignKey("games.id"))

    game: Mapped["Game"] = relationship("Game", back_populates="participants")
    stats: Mapped["ParticipantStat"] = relationship(
        "ParticipantStat", back_populates="participant"
    )

    team: Mapped["TeamDto"] = relationship(
        "TeamDto",
        primaryjoin="and_(Participant.game_id == TeamDto.game_id, Participant.team_id == TeamDto.team_id)",
        foreign_keys=[game_id, team_id],
        viewonly=True,
    )

    # Debug
    updated: Mapped[datetime] = mapped_column(DateTime, default=func.now(), onupdate=func.now())
    
    @hybrid_property
    def riot_name(self):
        return f"{self.riot_id_game_name}#{self.riot_id_tagline}"

    @riot_name.expression
    def riot_name(cls):
        return func.concat(cls.riot_id_game_name, "#", cls.riot_id_tagline)

    def __repr__(self):
        return f"{self.game_id}-{self.riot_name} (#{self.participant_id})"

    __table_args__ = (
        Index("idx_participants_game_id", "game_id"),
        Index("idx_participants_puuid", "puuid"),
        Index("idx_participants_puuid_queue", "puuid", postgresql_include=["game_id"]),
    )



from typing import Dict, Literal
from enum import StrEnum

headers = {
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.36",
    "Accept-Language": "en-US,en-CA;q=0.9,en;q=0.8,ru-RU;q=0.7,ru;q=0.6,ko-KR;q=0.5,ko;q=0.4",
    "Accept-Charset": "application/x-www-form-urlencoded; charset=UTF-8",
}

# The AMERICAS routing value serves NA, BR, LAN and LAS. The ASIA routing value serves KR and JP. The EUROPE routing value serves EUNE, EUW, ME1, TR and RU. The SEA routing value serves OCE, SG2, TW2 and VN2.
PLATFORM_ROUTING: Dict[str, str] = {
    "NA1": "americas",
    "BR1": "americas",
    "LA1": "americas",
    "LA2": "americas",
    "KR": "asia",
    "JP": "asia",
    "ME1": "europe",
    "EUN1": "europe",
    "EUW1": "europe",
    "TR1": "europe",
    "RU": "europe",
    "OC1": "sea",
    "SG2": "sea",
    "TW2": "sea",
    "VN2": "sea",
}

RoutingLiteral = Literal["americas", "asia", "europe", "sea"]
RegionLiteral = Literal[
    "NA1",
    "BR1",
    "LA1",
    "LA2",
    "KR",
    "JP",
    "ME1",
    "EUN1",
    "EUW1",
    "TR1",
    "RU",
    "OC1",
    "SG2",
    "TW2",
    "VN2",
]
QueueLiteral = Literal["RANKED_SOLO_5x5", "RANKED_FLEX_SR", "RANKED_FLEX_TT"]
DivisionLiteral = Literal["I", "II", "III", "IV"]
TierLiteral = Literal[
    "DIAMOND", "EMERALD", "PLATINUM", "GOLD", "SILVER", "BRONZE", "IRON"
]

REGIONS = StrEnum("Regions", [(_, _) for _ in PLATFORM_ROUTING.keys()])
QUEUES = StrEnum(
    "Queues", [(_, _) for _ in ["RANKED_SOLO_5x5", "RANKED_FLEX_SR", "RANKED_FLEX_TT"]]
)
DIVISIONS = StrEnum("Divisions", [(_, _) for _ in ["I", "II", "III", "IV"]])
TIERS = StrEnum(
    "Tiers",
    [
        (_, _)
        for _ in ["DIAMOND", "EMERALD", "PLATINUM", "GOLD", "SILVER", "BRONZE", "IRON"]
    ],
)


def parse_match_id(match_id: str) -> tuple[str, str]:
    split = match_id.split("_")
    if len(split) != 2:
        raise ValueError(f"Unable to parse {match_id}")
    if split[0] not in REGIONS:
        raise ValueError(f"Unknown region {split[0]} received")
    return split[0], split[1]



from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from ..models import Base


def get_session_factory(database_url: str, read_only: bool = False):
    engine = create_engine(database_url, pool_recycle=3600)
    if not read_only:
        Base.metadata.create_all(bind=engine)
    return sessionmaker(autocommit=False, autoflush=True, bind=engine)



from sqlalchemy import BigInteger, Integer, ForeignKey, DateTime, Boolean, Text, func
from sqlalchemy.orm import relationship, Mapped, mapped_column
from sqlalchemy.ext.hybrid import hybrid_property
from datetime import datetime
from .base import Base
from ..api.utils import REGIONS


class Account(Base):
    __tablename__ = "accounts"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    # This either the actual PUUID as per the Riot API OR Riot Esports API IDs
    puuid: Mapped[str] = mapped_column(Text, unique=True)
    name: Mapped[str] = mapped_column(Text)  # riotIdgameName
    tagline: Mapped[str] = mapped_column(Text)  # riotIdTagLine
    # Exceptions to the normal Riot API regions include TOURNAMENT (for TR accounts) and RIOT_LOL (for Riot Esports API)
    region: Mapped[str] = mapped_column(Text, nullable=False)
    # Last time / newest game was added to the database (used for match history ingestion)
    latest_game: Mapped[int | None] = mapped_column(BigInteger, nullable=True)
    updated: Mapped[datetime] = mapped_column(DateTime, default=func.now(), onupdate=func.now())
    # Player associated with the tracked account
    player_id: Mapped[int | None] = mapped_column(Integer, ForeignKey("players.id"))
    # Flag set for inactive tracked accounts
    skip_update: Mapped[bool] = mapped_column(Boolean, default=False)
    # We use an binary integer flag to save account details.
    account_details: Mapped[int] = mapped_column(Integer, default=0)

    @hybrid_property
    def solo_queue_account(self) -> bool:
        return self.region in REGIONS

    @solo_queue_account.expression
    def solo_queue_account(cls):
        return cls.region.in_(REGIONS)

    player: Mapped["Player"] = relationship("Player", back_populates="accounts")

    def __repr__(self) -> str:
        return f"{self.id} - {self.name}#{self.tagline}, region={self.region}, last_update={self.updated}"

    def __str__(self) -> str:
        return f"{self.name}#{self.tagline}"



from sqlalchemy import Text, Integer, DateTime, func
from sqlalchemy.orm import Mapped, mapped_column
from datetime import datetime
from .base import Base


class Champion(Base):
    __tablename__ = "champions"
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    name: Mapped[str] = mapped_column(Text, unique=True)
    # alias (communitydragon) is the same as id (ddragon)
    alias: Mapped[str] = mapped_column(Text, unique=True)

    updated: Mapped[datetime] = mapped_column(DateTime, default=func.now(), onupdate=func.now())



from sqlalchemy import (
    Boolean,
    ForeignKey,
    Integer,
    DateTime,
    Float,
    Index,
    func,
    case,
    type_coerce,
    select,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import Mapped, mapped_column, MappedColumn, relationship
from datetime import datetime
from .base import Base


class ParticipantStat(Base):
    __tablename__ = "participant_stats"
    ### Original MatchV5-named Columns
    kills: Mapped[int] = mapped_column(Integer)
    deaths: Mapped[int] = mapped_column(Integer)
    assists: Mapped[int] = mapped_column(Integer)
    total_minions_killed: Mapped[int] = mapped_column(Integer) # Total *lane* minions killed.
    neutral_minions_killed: Mapped[int] = mapped_column(Integer) # Total *jungle* minions killed.

    # 0 if the item slot is empty. Should be a foreign key in the future.
    item_0: Mapped[int] = mapped_column(Integer, default=0)
    item_1: Mapped[int] = mapped_column(Integer, default=0)
    item_2: Mapped[int] = mapped_column(Integer, default=0)
    item_3: Mapped[int] = mapped_column(Integer, default=0)
    item_4: Mapped[int] = mapped_column(Integer, default=0)
    item_5: Mapped[int] = mapped_column(Integer, default=0)
    item_6: Mapped[int] = mapped_column(Integer, default=0)  # Trinket

    # TODO: Should also be a foreign key
    summoner_1_id: Mapped[int] = mapped_column(Integer)
    summoner_2_id: Mapped[int] = mapped_column(Integer)

    total_damage_dealt: Mapped[int | None] = mapped_column(Integer, nullable=True)
    physical_damage_dealt: Mapped[int | None] = mapped_column(Integer, nullable=True)
    magic_damage_dealt: Mapped[int | None] = mapped_column(Integer, nullable=True)
    true_damage_dealt: Mapped[int | None] = mapped_column(Integer, nullable=True)

    total_damage_dealt_to_champions: Mapped[int | None] = mapped_column(Integer, nullable=True)
    physical_damage_dealt_to_champions: Mapped[int | None] = mapped_column(Integer, nullable=True)
    magic_damage_dealt_to_champions: Mapped[int | None] = mapped_column(Integer, nullable=True)
    true_damage_dealt_to_champions: Mapped[int | None] = mapped_column(Integer, nullable=True)

    total_damage_taken: Mapped[int | None] = mapped_column(Integer, nullable=True)
    physical_damage_taken: Mapped[int | None] = mapped_column(Integer, nullable=True)
    magic_damage_taken: Mapped[int | None] = mapped_column(Integer, nullable=True)
    true_damage_taken: Mapped[int | None] = mapped_column(Integer, nullable=True)

    damage_self_mitigated: Mapped[int | None] = mapped_column(Integer, nullable=True)
    total_damage_shielded_on_teammates: Mapped[int | None] = mapped_column(Integer, nullable=True)

    damage_dealt_to_buildings: Mapped[int | None] = mapped_column(Integer, nullable=True)
    damage_dealt_to_turrets: Mapped[int | None] = mapped_column(Integer, nullable=True)
    total_damage_dealt_to_champions: Mapped[int | None] = mapped_column(Integer, nullable=True)

    wards_placed: Mapped[int | None] = mapped_column(Integer, nullable=True)
    wards_killed: Mapped[int | None] = mapped_column(Integer, nullable=True)
    vision_score: Mapped[int | None] = mapped_column(Integer, nullable=True)

    total_heals_on_teammates: Mapped[int | None] = mapped_column(Integer, nullable=True)
    total_ally_jungle_minions_killed: Mapped[int | None] = mapped_column(Integer, nullable=True)
    total_enemy_jungle_minions_killed: Mapped[int | None] = mapped_column(Integer, nullable=True)

    champ_level: Mapped[int | None] = mapped_column(Integer, nullable=True)
    champ_experience: Mapped[int | None] = mapped_column(Integer, nullable=True)

    basic_pings: Mapped[int | None] = mapped_column(Integer, nullable=True)
    command_pings: Mapped[int | None] = mapped_column(Integer, nullable=True)
    danger_pings: Mapped[int | None] = mapped_column(Integer, nullable=True)
    get_back_pings: Mapped[int | None] = mapped_column(Integer, nullable=True)
    retreat_pings: Mapped[int | None] = mapped_column(Integer, nullable=True)
    on_my_way_pings: Mapped[int | None] = mapped_column(Integer, nullable=True)
    assist_me_pings: Mapped[int | None] = mapped_column(Integer, nullable=True)
    enemy_missing_pings: Mapped[int | None] = mapped_column(Integer, nullable=True)
    push_pings: Mapped[int | None] = mapped_column(Integer, nullable=True)
    all_in_pings: Mapped[int | None] = mapped_column(Integer, nullable=True)
    hold_pings: Mapped[int | None] = mapped_column(Integer, nullable=True)
    vision_cleared_pings: Mapped[int | None] = mapped_column(Integer, nullable=True)
    enemy_vision_pings: Mapped[int | None] = mapped_column(Integer, nullable=True)
    need_vision_pings: Mapped[int | None] = mapped_column(Integer, nullable=True)

    # The columns defined above can be automatically parsed for MatchV5 inputs
    PARTICIPANT_STAT_DTO = [
        name for name, value in locals().items() if isinstance(value, MappedColumn)
    ]

    ### The following columns require additional processing for MatchV5 inputs
    total_time_CC_dealt: Mapped[int | None] = mapped_column(Integer, nullable=True)
    time_CC_ing_others:  Mapped[int | None] = mapped_column(Integer, nullable=True)
    current_gold: Mapped[int | None] = mapped_column(Integer, nullable=True)
    total_gold: Mapped[int | None] = mapped_column(Integer, nullable=True)
    perks: Mapped[dict | None] = mapped_column(JSONB)

    ### The following columns must stay nullable
    # stats_update specific column
    position_x: Mapped[int | None] = mapped_column(Integer, nullable=True)
    position_y: Mapped[int | None] = mapped_column(Integer, nullable=True) # also refered to as position z
    alive: Mapped[bool | None] = mapped_column(Boolean, nullable=True)
    respawn_time: Mapped[int | None] = mapped_column(Integer, nullable=True)

    # This JSONB column to stores the rest of the Match-V5 data for future use
    source_data = mapped_column(JSONB)

    ### Identification Columns
    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    # game_time is NULL for match_v5 summaries
    game_time: Mapped[int | None] = mapped_column(Integer, nullable=True)
    participant_id: Mapped[int] = mapped_column(
        Integer, ForeignKey("participants.id")
    )

    participant: Mapped["Participant"] = relationship(
        "Participant", back_populates="stats"
    )

    @classmethod
    def _game_duration_minutes(cls):
        from .participant import Participant
        from .game import Game

        return (
            select(Game.game_duration / 60)
            .join(Participant, Participant.game_id == Game.id)
            .where(Participant.id == cls.participant_id)
            .scalar_subquery()
        )

    @hybrid_property
    def cs(self) -> int:
        return self.total_minions_killed + self.neutral_minions_killed

    @cs.expression
    @classmethod
    def cs(cls):
        return type_coerce(
            cls.total_minions_killed + cls.neutral_minions_killed, Integer
        )

    @hybrid_property
    def cspm(self) -> float | None:
        return self.cs / (self.participant.game.game_duration / 60)

    @cspm.expression
    @classmethod
    def cspm(cls):
        return type_coerce(cls.cs / (cls._game_duration_minutes()), Float)

    @hybrid_property
    def kda(self) -> float:
        if self.deaths > 0:
            return (self.kills + self.assists) / self.deaths
        return self.kills + self.assists

    @kda.expression
    @classmethod
    def kda(cls):
        return case(
            (cls.deaths > 0, (cls.kills + cls.assists) / cls.deaths),
            else_=(cls.kills + cls.assists),
        )

    __table_args__ = (
        Index("idx_participant_stats_participant_time", "participant_id", "game_time"),
        Index("idx_participant_stats_participant_id", "participant_id"),
        Index("idx_participant_stats_game_time", "game_time"),
    )

    # Debug
    updated: Mapped[datetime] = mapped_column(DateTime, default=func.now(), onupdate=func.now())



from .account_v1 import get_account_by_puuid, get_account_by_riot_id
from .ddragon import get_ddragon_versions
from .league_v4 import (
    get_masters,
    get_grandmasters,
    get_challengers,
    get_league_entries_by_puuid,
    get_league_entries_by_summoner_id,
    get_all_league_entries,
    get_league_entries_by_league_id,
)
from .match_v5 import get_available_matches, get_match_by_id, get_match_history
from .spectator_v5 import get_active_games, get_featured_games
from .summoner_v4 import (
    get_summoner_by_account_id,
    get_summoner_by_puuid,
    get_summoner_by_summoner_id,
)
from .utils import parse_match_id, REGIONS, QUEUES, DIVISIONS, TIERS


__all__ = [
    "get_account_by_puuid",
    "get_account_by_riot_id",
    "get_ddragon_versions",
    "get_masters",
    "get_grandmasters",
    "get_challengers",
    "get_league_entries_by_puuid",
    "get_league_entries_by_summoner_id",
    "get_all_league_entries",
    "get_league_entries_by_league_id",
    "get_available_matches",
    "get_match_by_id",
    "get_match_history",
    "get_active_games",
    "get_featured_games",
    "get_summoner_by_account_id",
    "get_summoner_by_puuid",
    "get_summoner_by_summoner_id",
    "parse_match_id",
    "REGIONS",
    "QUEUES",
    "DIVISIONS",
    "TIERS",
]



from datetime import datetime, timedelta
from sqlalchemy import and_, select
from sqlalchemy.orm import Session as _Session
from tqdm import tqdm
from ..api import get_match_history, get_match_by_id
from ..api.account_v1 import get_account_by_puuid
from ..models import Player, Game, Participant, Account, TeamDto, ParticipantStat
from ..utils import SEASON_START, snake_to_camel


def update_account_names(session: _Session, API_KEY: str, days: int = 7):
    """Updates player account names/taglines"""
    update_delta = datetime.now() - timedelta(days=days)
    accounts_to_update = list(
        session.scalars(
            select(Account).where(
                and_(
                    Account.updated < update_delta,
                    Account.solo_queue_account == True,
                    Account.skip_update == False,
                )
            )
        )
    )
    if len(accounts_to_update) == 0:
        print("All accounts are up to date!")
        return
    for account in tqdm(accounts_to_update):
        account_details = get_account_by_puuid(account.puuid, API_KEY)
        account_details = account_details.json()
        account.name = account_details["gameName"]
        account.tagline = account_details["tagLine"]
        account.updated = datetime.now()
    try:
        session.commit()
    except Exception as e:
        print(f"Something went wrong updating accounts: {str(e)}")
        session.rollback()


def insert_match_history(
    session: _Session,
    player: Player,
    API_KEY: str,
    start_time: int = SEASON_START,
    start_latest: bool = True,
    queue_id: int = 420,
):
    existing_ids = set(session.scalars(select(Game.id)).all())
    for account in player.accounts or []:
        if not account.solo_queue_account or account.skip_update:
            continue
        print(f"Updating match history for {str(account)}")

        if start_latest:
            latest_game_timestamp = session.scalar(
                select(Game.game_end_timestamp)
                .join(Participant, Participant.game_id == Game.id)
                .where(
                    and_(Participant.puuid == account.puuid, Game.queue_id == queue_id)
                )
                .order_by(Game.game_end_timestamp.desc())
                .limit(1)
            )
            if latest_game_timestamp is None:
                start_time = SEASON_START
            else:
                start_time = int(latest_game_timestamp.timestamp())
        match_ids = get_match_history(
            account.puuid,
            account.region,
            API_KEY,
            startTime=start_time,
            queue=queue_id,
        )
        new_match_ids = set(match_ids) - existing_ids

        if len(new_match_ids) == 0:
            print("All up to date!")
            continue

        for match_id in tqdm(new_match_ids):
            try:
                game_data = get_match_by_id(match_id, API_KEY).json()["info"]
                process_match(session, match_id, game_data)
                existing_ids.add(match_id)
            except Exception as e:
                print(f"Failed to process match {match_id}: {str(e)}")
        session.commit()


def process_match(
    session: _Session,
    game_id: str,
    game_data: dict,
    game_args: dict = {},
) -> int | None:
    """Function to process game data from MatchV5-like sources. If the game already exists, processing will be skipped.

    Args:
        session: ORM Session object
        game_id: Riot's Game ID number
        match_data: MatchV5-like Dictionary to be processed
        game_args: Additional fields to be passed to the game object
    """
    if session.query(Game).filter(Game.id == game_id).one_or_none() is not None:
        return

    if len(game_data.get("participants", [])) > 0:
        early_surrender = game_data["participants"][0].get(
            "gameEndedInEarlySurrender", False
        )
        surrender = game_data["participants"][0].get("gameEndedInSurrender", False)
    else:
        early_surrender = False
        surrender = False

    game = Game(
        **{k: game_data.get(snake_to_camel(k)) for k in Game.INFO_DTO},
        **{k: datetime.fromtimestamp(game_data[snake_to_camel(k)] / 1000) for k in Game.TIMESTAMPS},
        **{"id": game_id},
        **game_args,
        game_ended_in_early_surrender=early_surrender,
        game_ended_in_surrender=surrender,
    )

    session.add(game)
    session.flush() # We need to flush to ensure the game id exists in the database.

    for team in game_data["teams"]:
        teamDto = TeamDto(
            game_id=game.id,
            bans=team["bans"],
            objectives=team["objectives"],
            team_id=team["teamId"],
            win=team["win"],
        )
        session.add(teamDto)

    # For some reason, zero duration game's are permitted by GRID. We can not process the match data for these games
    if game.game_duration and game.game_duration > 0:
        participant_stats = []
        game_participants = game_data["participants"]
        for participant in game_participants:
            # First we extract the columns defined in the model and the DTOs we're storing
            participant_dto_extraction = {
                k: participant[snake_to_camel(k)] for k in Participant.PARTICIPANT_DTO
            }
            participant_stat_extraction = {
                k: participant[snake_to_camel(k)]
                for k in ParticipantStat.PARTICIPANT_STAT_DTO
            }

            # Handle special fields that require extra processing
            special_fields = {}
            special_fields["perks"] = participant.get("perks")
            special_fields["total_time_CC_dealt"] = int(participant.get("totalTimeCCDealt", 0))
            special_fields["time_CC_ing_others"] = int(participant.get("timeCCingOthers", 0))
            special_fields["total_gold"] = int(participant.get("goldEarned", 0))
            special_fields["current_gold"] = int(participant.get("goldEarned", 0) - participant.get("goldSpent", 0))

            source_data = dict(participant)

            keys_to_remove = (
                ParticipantStat.PARTICIPANT_STAT_DTO +
                Participant.PARTICIPANT_DTO +
                ["perks", "totalTimeCCDealt", "timeCCingOthers", "goldEarned", "goldSpent"]
            )
            for key in keys_to_remove:
                camel_key = snake_to_camel(key) if key in ParticipantStat.PARTICIPANT_STAT_DTO + Participant.PARTICIPANT_DTO else key
                source_data.pop(camel_key, None)

            new_participant = Participant(game_id=game.id, **participant_dto_extraction)
            session.add(new_participant)
            session.flush()
            participant_stats.append(
                ParticipantStat(
                    participant_id=new_participant.id,
                    **participant_stat_extraction,
                    **special_fields,
                    source_data=source_data,
                )
            )

        session.add_all(participant_stats)
    try:
        session.commit()
    except Exception as e:
        session.rollback()
        print(f"Something went wrong ingesting match {game_id}: {str(e)}")



"""
A lightweight, decorator-based approach to Riot API rate limiting.
Created with Claude 3.7
"""

import time
import functools
from typing import Dict, Callable, Any, Optional
import threading


class RiotRateLimiter:
    """Simple rate limiter for Riot API based on response headers."""

    def __init__(self):
        # Store rate limit info by endpoint key (method+region)
        self.limits = {}
        # Add thread lock for thread safety
        self.lock = threading.Lock()

    def update_limits(self, endpoint_key: str, headers: Dict[str, str]) -> None:
        """Update rate limit information from response headers."""
        with self.lock:
            # Track when we should retry if rate limited
            retry_after = headers.get("Retry-After")
            if retry_after:
                self.limits[endpoint_key] = {
                    "retry_after": int(retry_after) + time.time()
                }
                return

            # Parse app rate limit headers
            app_limit = headers.get("X-App-Rate-Limit")
            app_count = headers.get("X-App-Rate-Limit-Count")

            # Parse method rate limit headers
            method_limit = headers.get("X-Method-Rate-Limit")
            method_count = headers.get("X-Method-Rate-Limit-Count")

            # Only update if we have both values
            if not (app_limit and app_count and method_limit and method_count):
                return

            # Parse all time windows from app rate limit
            app_limits = {}
            app_counts = {}

            for limit_part, count_part in zip(
                app_limit.split(","), app_count.split(",")
            ):
                limit_val, window = limit_part.split(":")
                count_val, _ = count_part.split(":")
                window = int(window)
                app_limits[window] = int(limit_val)
                app_counts[window] = int(count_val)

            # Parse all time windows from method rate limit
            method_limits = {}
            method_counts = {}

            for limit_part, count_part in zip(
                method_limit.split(","), method_count.split(",")
            ):
                limit_val, window = limit_part.split(":")
                count_val, _ = count_part.split(":")
                window = int(window)
                method_limits[window] = int(limit_val)
                method_counts[window] = int(count_val)

            # Store the limit data with expiry time
            current_time = time.time()
            limit_data = {
                "app_limits": app_limits,
                "app_counts": app_counts,
                "app_reset": current_time + max(app_limits.keys()),
                "method_limits": method_limits,
                "method_counts": method_counts,
                "method_reset": current_time + max(method_limits.keys()),
                "last_updated": current_time,
            }

            # Update or set the limits for this endpoint
            self.limits[endpoint_key] = limit_data

    def should_wait(self, endpoint_key: str) -> float:
        """Determine if/how long we should wait before making another request."""
        with self.lock:
            if endpoint_key not in self.limits:
                return 0

            limit_data = self.limits[endpoint_key]
            current_time = time.time()

            # Handle explicit retry-after
            if "retry_after" in limit_data:
                wait_time = limit_data["retry_after"] - current_time
                if wait_time <= 0:
                    # Clear the retry after
                    del self.limits[endpoint_key]["retry_after"]
                    return 0
                return wait_time

            # If we don't have limit data yet, don't wait
            if not ("app_limits" in limit_data and "method_limits" in limit_data):
                return 0

            # Check all app limit windows
            app_wait = 0
            for window, limit in limit_data["app_limits"].items():
                count = limit_data["app_counts"].get(window, 0)
                usage_ratio = count / limit

                # If we're at more than 80% of capacity, calculate wait time
                if usage_ratio > 0.8:
                    # Estimate time until reset based on window size
                    time_elapsed = current_time - limit_data["last_updated"]
                    time_remaining = max(0, window - time_elapsed)
                    wait = time_remaining * (usage_ratio - 0.7)  # Progressive wait
                    app_wait = max(app_wait, wait)

            # Check all method limit windows
            method_wait = 0
            for window, limit in limit_data["method_limits"].items():
                count = limit_data["method_counts"].get(window, 0)
                usage_ratio = count / limit

                # If we're at more than 80% of capacity, calculate wait time
                if usage_ratio > 0.8:
                    # Estimate time until reset based on window size
                    time_elapsed = current_time - limit_data["last_updated"]
                    time_remaining = max(0, window - time_elapsed)
                    wait = time_remaining * (usage_ratio - 0.7)  # Progressive wait
                    method_wait = max(method_wait, wait)

            # Take the larger of the two wait times
            return max(app_wait, method_wait)


# Create a single shared rate limiter
_rate_limiter = RiotRateLimiter()


def riot_api_limiter(endpoint_key: Optional[str] = None):
    """
    Decorator that handles Riot API rate limiting based on response headers.

    Args:
        endpoint_key: Optional key to identify this endpoint. If not provided,
                     the function name will be used.
    """

    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Get the endpoint key
            key = endpoint_key or func.__name__

            # Check region if it's in the kwargs
            if "region" in kwargs:
                key = f"{key}_{kwargs['region']}"

            # Wait if needed
            wait_time = _rate_limiter.should_wait(key)
            if wait_time > 0:
                time.sleep(wait_time)

            # Call the original function
            response = func(*args, **kwargs)

            # Update rate limits from the response
            if hasattr(response, "headers"):
                _rate_limiter.update_limits(key, response.headers)

            # Handle rate limiting response (429)
            if hasattr(response, "status_code") and response.status_code == 429:
                retry_after = int(response.headers.get("Retry-After", 1))
                time.sleep(retry_after)
                # Retry the request
                return func(*args, **kwargs)

            return response

        return wrapper

    return decorator


def print_rate_limit_headers(response):
    """Helper function to print rate limit headers from a response."""
    print("\nRate Limit Headers:")
    for key, value in response.headers.items():
        if "rate" in key.lower() or "retry" in key.lower():
            print(f"{key}: {value}")



import requests as r
from requests import Response
from .utils import headers, RegionLiteral
from ..rate_limiter import riot_api_limiter


@riot_api_limiter(endpoint_key="spectator_v5_active_games")
def get_active_games(
    puuid: str, region: RegionLiteral, api_key: str, **kwargs
) -> Response:
    _headers = {"X-Riot-Token": api_key, **headers}
    response = r.get(
        f"https://{region.lower()}.api.riotgames.com/lol/spectator/v5/active-games/by-summoner/{puuid}",
        headers=_headers,
        params=kwargs,
    )
    return response


@riot_api_limiter(endpoint_key="spectator_v5_featured_games")
def get_featured_games(region: RegionLiteral, api_key: str, **kwargs) -> Response:
    _headers = {"X-Riot-Token": api_key, **headers}
    response = r.get(
        f"https://{region.lower()}.api.riotgames.com/lol/spectator/v5/featured-games",
        headers=_headers,
        params=kwargs,
    )
    return response



import requests as r
from requests import Response
from .utils import headers, RoutingLiteral
from ..rate_limiter import riot_api_limiter


@riot_api_limiter(endpoint_key="account_v1_by_riot_id")
def get_account_by_riot_id(
    game_name: str, tag_line: str, api_key: str, routing: RoutingLiteral = "americas"
) -> Response:
    _headers = {"X-Riot-Token": api_key, **headers}
    # There are three routing values for account-v1; americas, asia, and europe. You can query for any account in any region. We recommend using the nearest cluster.
    response = r.get(
        f"https://{routing}.api.riotgames.com/riot/account/v1/accounts/by-riot-id/{game_name}/{tag_line}",
        headers=_headers,
    )
    return response


@riot_api_limiter(endpoint_key="account_v1_by_puuid")
def get_account_by_puuid(
    puuid: str, api_key: str, routing: RoutingLiteral = "americas"
) -> Response:
    _headers = {"X-Riot-Token": api_key, **headers}
    response = r.get(
        f"https://{routing}.api.riotgames.com/riot/account/v1/accounts/by-puuid/{puuid}",
        headers=_headers,
    )
    return response



from .esports_ingest import (
    create_account,
    create_player,
    create_missing_accounts,
    parse_game_participants,
    get_common_prefix_length
)
from .ingest_match import (
    insert_match_history,
    update_account_names,
    process_match,
)

__all__ = [
    "create_account",
    "create_player",
    "create_missing_accounts",
    "parse_game_participants",
    "get_common_prefix_length",
    "insert_match_history",
    "update_account_names",
    "process_match",
]



import re


# https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
def camel_to_snake(s: str) -> str:
    pattern = re.compile(r"(?<!^)(?=[A-Z0-9])")
    return pattern.sub("_", s).lower()


def snake_to_camel(s: str) -> str:
    words = s.split("_")
    return words[0].lower() + "".join(word.title() for word in words[1:])



import requests as r
from requests import Response
from .utils import headers, RegionLiteral
from ..rate_limiter import riot_api_limiter


@riot_api_limiter(endpoint_key="summoner_v4_by_account_id")
def get_summoner_by_account_id(
    account_id: str, region: RegionLiteral, api_key: str, **kwargs
) -> Response:
    _headers = {"X-Riot-Token": api_key, **headers}
    response = r.get(
        f"https://{region.lower()}.api.riotgames.com/lol/summoner/v4/summoners/by-account/{account_id}",
        headers=_headers,
        params=kwargs,
    )
    return response


@riot_api_limiter(endpoint_key="summoner_v4_by_puuid")
def get_summoner_by_puuid(
    puuid: str, region: RegionLiteral, api_key: str, **kwargs
) -> Response:
    _headers = {"X-Riot-Token": api_key, **headers}
    response = r.get(
        f"https://{region.lower()}.api.riotgames.com/lol/summoner/v4/summoners/by-puuid/{puuid}",
        headers=_headers,
        params=kwargs,
    )
    return response


@riot_api_limiter(endpoint_key="summoner_v4_by_summoner_id")
def get_summoner_by_summoner_id(
    summoner_id: str, region: RegionLiteral, api_key: str, **kwargs
) -> Response:
    _headers = {"X-Riot-Token": api_key, **headers}
    response = r.get(
        f"https://{region.lower()}.api.riotgames.com/lol/summoner/v4/summoners/{summoner_id}",
        headers=_headers,
        params=kwargs,
    )
    return response



from sqlalchemy import Integer, Text, ForeignKey, DateTime, func, Index
from datetime import datetime
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import Mapped, mapped_column
from .base import Base


class GameEvent(Base):
    __tablename__ = "game_events"
    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    game_id: Mapped[str] = mapped_column(Text, ForeignKey("games.id"))
    # rfc461Schema (identifies the event type)
    schema: Mapped[str] = mapped_column(Text)
    sequence_index: Mapped[int] = mapped_column(Integer)
    # Missing from 'champ_select', 'game_info' type events
    game_time: Mapped[int] = mapped_column(Integer)

    # The rest of the event data is stored here
    source_data = mapped_column(JSONB)

    # Debug
    updated: Mapped[datetime] = mapped_column(DateTime, default=func.now(), onupdate=func.now())

    __table_args__ = (
        Index("idx_game_events_game_id", "game_id"),
        Index("idx_game_events_game_id_game_time", "game_id", "game_time"),
        Index("idx_game_events_game_id_schema", "game_id", "schema"),
        Index(
            "idx_game_events_game_id_schema_seq", "game_id", "schema", "sequence_index"
        ),
        Index("idx_game_events_game_id_schema_time", "game_id", "schema", "game_time"),
        Index("idx_game_events_schema", "schema"),
    )



from sqlalchemy import ForeignKey, Text, Integer, DateTime, func
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import Mapped, mapped_column
from datetime import datetime
from .base import Base
from ..utils.constants import SeriesStatus, SeriesType

class Series(Base):
    __tablename__ = "series"
    ###          1:1 GRID          ###
    id: Mapped[str] = mapped_column(Text, primary_key=True)
    type: Mapped[SeriesType] # GRID SeriesType
    # Some of the older series have null start times when data is not available
    scheduled_start_time: Mapped[datetime | None] = mapped_column(DateTime, nullable=True)
    format: Mapped[int] = mapped_column(Integer) # Best of number
    tournament_id: Mapped[str] = mapped_column(Text, ForeignKey("tournaments.id"))
    external_links = mapped_column(JSONB)

    ###     Processing status      ###
    status: Mapped[SeriesStatus]

    ###           Debug            ###
    updated: Mapped[datetime] = mapped_column(DateTime, default=func.now(), onupdate=func.now())